`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    19:42:47 11/26/2011 
// Design Name: 
// Module Name:    alu 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module alu(
		oe,
		opcode,
		in1,
		in2,
		carry,
		out,
		flags
	);

`define PASS1		0
`define PASS2		1
`define ADD 		2
`define ADC 		3
`define INC 		4
`define SUB			5
`define SBB			6
`define DEC			7
`define NEG			8
`define AND			9
`define OR			10
`define XOR			11
`define NOT			12
`define SHL			13
`define SHR			14
`define SAR			15

parameter width = 16;
parameter flags_width = 5;

input							oe;
input [3:0]					opcode;
input	[width-1 : 0]		in1;
input	[width-1 : 0]		in2;
input							carry;
output[width-1 : 0]		out;
output[flags_width-1 : 0]	flags;	// P, S, Z, O, C

reg	[width-1 : 0]		result;
reg							p, s, z, o, c;

always @(*) begin
	case(opcode)
		`PASS1: begin
				result = in1;
				c = 0;
				o = 0;
			end
		
		`PASS2: begin
				result = in2;
				c = 0;
				o = 0;
			end
		
		`ADD: begin
				{c, result} = in1 + in2;
				o = (in1[width-1] == in2[width-1]) && (in1[width-1] != result[width-1]);
			end
		
		`ADC: begin
				{c, result} = in1 + in2 + carry;
				o = (in1[width-1] == in2[width-1]) && (in1[width-1] != result[width-1]);
			end
		
		`INC: begin
				{c, result} = in1 + 1;
				o = (in1[width-1] == 0) && (in1[width-1] != result[width-1]);
			end
		
		`SUB: begin
				{c, result} = in1 - in2;
				o = (in1[width-1] != in2[width-1]) && (in1[width-1] != result[width-1]);
			end
		
		`SBB: begin
				{c, result} = in1 - in2 - carry;
				o = (in1[width-1] != in2[width-1]) && (in1[width-1] != result[width-1]);
			end
		
		`DEC: begin
				{c, result} = in1 - 1;
				o = (in1[width-1] != 0) && (in1[width-1] != result[width-1]);
			end
		
		`NEG: begin
				result = -in1;
				c = 0;
				o = in1[width-1] == result[width-1];
			end
		
		`AND: begin
				result = in1 & in2;
				c = 0;
				o = 0;
			end
		
		`OR: begin
				result = in1 | in2;
				c = 0;
				o = 0;
			end
		
		`XOR: begin
				result = in1 ^ in2;
				c = 0;
				o = 0;
			end
		
		`NOT: begin
				result = ~in1;
				c = 0;
				o = 0;
			end
		
		`SHL: begin
				result = in1 << 1;
				c = 0;
				o = 0;
			end
		
		`SHR: begin
				result = in1 >> 1;
				c = 0;
				o = 0;
			end
		
		`SAR: begin
				result = in1 >>> 1;
				c = 0;
				o = 0;
			end
		
		default: begin
				result = 0;
				c = 0;
				o = 0;
			end
	endcase
	
	z = result == 0;
	s = result[width-1];
	p = ^result;
end

assign out = oe ? result : 0;
assign flags = {p, s, z, o, c};

endmodule
