`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    20:14:45 11/26/2011 
// Design Name: 
// Module Name:    uc 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
module uc(
		clk,
		rst,
		ri,
		ind,
		regs_addr,
		regs_oe,
		regs_we,
		alu_oe,
		alu_carry,
		alu_opcode,
		ram_oe,
		ram_we,
		cp_oe,
		cp_we,
		ind_sel,
		ind_oe,
		ind_we,
		am_oe,
		am_we,
		t1_oe,
		t1_we,
		t2_oe,
		t2_we,
		ri_oe,
		ri_we
	);

parameter word_width = 16;

`define PASS1		0
`define PASS2		1
`define ADD 		2
`define ADC 		3
`define INC 		4
`define SUB			5
`define SBB			6
`define DEC			7
`define NEG			8
`define AND			9
`define OR			10
`define XOR			11
`define NOT			12
`define SHL			13
`define SHR			14
`define SAR			15

`define RA			0
`define RB			1
`define RC			2
`define IS			3
`define XA			4
`define XB			5
`define BA			6
`define BB			7

input								clk;
input								rst;
input [word_width-1 : 0]	ri;
input [word_width-1 : 0]	ind;
output reg						alu_oe;
output reg						alu_carry;
output reg[3 : 0]				alu_opcode;
output reg						ram_oe;
output reg						ram_we;
output reg[2 : 0]				regs_addr;
output reg						regs_oe;
output reg						regs_we;
output reg						cp_oe;
output reg						cp_we;
output reg						ind_sel;		// controls IND register input (0 = bus, 1 = alu flags)
output reg						ind_oe;
output reg						ind_we;
output reg						am_oe;
output reg						am_we;
output reg						t1_oe;
output reg						t1_we;
output reg						t2_oe;
output reg						t2_we;
output reg						ri_oe;			// controls RI register output which generates the offset for Jcond instructions
output reg						ri_we;

wire [1:0]						mod;
wire [2:0]						rg;
wire [2:0]						rm;
wire								d;

assign mod = {ri[8], ri[9]};
assign rg = {ri[10], ri[11], ri[12]};
assign rm = {ri[13], ri[14], ri[15]};
assign d = ri[7];

parameter state_width = 32;

`define reset	0								// reset state
`define fetch	10								// load instruction to instruction register
`define decode	20								// analyze loaded instruction
`define push	30								// store word to stack
`define dec_is	35								// decrement stack pointer
`define pop		40								// load word from stack
`define inc_is	45								// increment stack pointer
`define inc_cp	50								// increment program counter
`define load_rm_reg 60							// load operand from RM when MOD == 11
`define load_rm_mem 65							// load operand from RM when MOD != 11
`define load_rg 70								// load operand from REG
`define exec 80									// execute operation and store result
`define addr_sum 90								// load address when MOD == 00 and RI[13] == 0
`define addr_reg 100							// load address when MOD == 00 and RI[13] == 1

reg	[state_width-1 : 0]	state = `reset;

// FSM next state logic
always @(posedge clk) begin
	state <= `reset;
	
	if(!rst) begin
		case(state)
			`reset: begin
					state <= `fetch;
				end
			
			`fetch: begin
					state <= `fetch + 1;
				end
			
			`fetch + 'd1: begin
					state <= `fetch + 2;
				end
			
			`fetch + 'd2: begin
					state <= `decode;
				end
			
			`decode: begin
					if(mod == 2'b00)
						; //TODO goto compute and load operand address
					else if(mod == 2'b11)
						if(ri[2:0] == 3'b010)
							state <= `load_rm_reg;
						else if(ri[6:0] == 7'b0100000)
							state <= `dec_is;
						else if(ri[6:0] == 7'b1100000)
							state <= `pop;
				end
			
			`push: begin
					state <= `push + 1;
				end
			
			`push + 'd1: begin
					state <= `push + 2;
				end
			
			`push + 'd2: begin
					state <= `inc_cp;
				end
			
			`dec_is: begin
					state <= `dec_is + 1;
				end
			
			`dec_is + 'd1: begin
					state <= `push;
				end
			
			`pop: begin
					state <= `pop + 1;
				end
				
			`pop + 'd1: begin
					state <= `pop + 2;
				end
			
			`pop + 'd2: begin
					state <= `inc_is;
				end
			
			`inc_is: begin
					state <= `inc_is + 1;
				end
			
			`inc_is + 'd1: begin
					state <= `inc_cp;
				end
			
			`inc_cp: begin
					state <= `inc_cp + 1;
				end
			
			`inc_cp + 'd1: begin
					state <= `fetch;
				end
			
			`load_rm_reg: begin
					state <= `load_rg;
				end
			
			`load_rm_mem: begin
					state <= `load_rm_mem + 1;
				end
			
			`load_rm_mem + 'd1: begin
					state <= `load_rg;
				end
			
			`load_rg: begin
					state <= `exec;
				end
			
			`exec: begin
					if(mod == 2'b11 || d == 1)
						state <= `inc_cp;
					else
						state <= `exec + 1;
				end
			
			`exec + 'd1: begin
					state <= `inc_cp;
				end
			
			`addr_sum: begin
					state <= `addr_sum + 1;
				end
			
			`addr_sum + 'd1: begin
					state <= `addr_sum + 2;
				end
			
			`addr_sum + 'd2: begin
					// TODO goto load operand from computed address
				end
			
			`addr_reg: begin
					// TODO goto load operand from computed address
				end
			
			default: ;
		endcase
	end
end

// FSM output logic
always @(*) begin
	alu_oe = 0;
	alu_carry = 0;
	alu_opcode = 0;
	ram_oe = 0;
	ram_we = 0;
	regs_addr = 0;
	regs_oe = 0;
	regs_we = 0;
	cp_oe = 0;
	cp_we = 0;
	ind_sel = 0;
	ind_oe = 0;
	ind_we = 0;
	am_oe = 0;
	am_we = 0;
	t1_oe = 0;
	t1_we = 0;
	t2_oe = 0;
	t2_we = 0;
	ri_oe = 0;
	ri_we = 0;
	
	case(state)
		`fetch: begin
				cp_oe = 1;
				am_we = 1;
			end
		
		`fetch + 'd1: begin
				am_oe = 1;
			end
		
		`fetch + 'd2: begin
				ram_oe = 1;
				ri_we = 1;
			end
		
		`push: begin
				am_we = 1;
				regs_addr = `IS;
				regs_oe = 1;
			end
		
		`push + 'd1: begin
				am_oe = 1;
				ram_we = 1;
				regs_addr = rm;
				regs_oe = 1;
			end
		
		`push + 'd2: begin
			end
		
		`dec_is: begin
				t1_we = 1;
				regs_addr = `IS;
				regs_oe = 1;
			end
		
		`dec_is + 'd1: begin
				regs_addr = `IS;
				regs_we = 1;
				t1_oe = 1;
				alu_opcode = `DEC;
				alu_oe = 1;
			end
		
		`pop: begin
				am_we = 1;
				regs_addr = `IS;
				regs_oe = 1;
			end
		
		`pop + 'd1: begin
				am_oe = 1;
			end
		
		`pop + 'd2: begin
				regs_addr = rm;
				regs_we = 1;
				ram_oe = 1;
			end
		
		`inc_is: begin
				t1_we = 1;
				regs_addr = `IS;
				regs_oe = 1;
			end
		
		`inc_is + 'd1: begin
				regs_addr = `IS;
				regs_we = 1;
				t1_oe = 1;
				alu_opcode = `INC;
				alu_oe = 1;
			end
		
		`inc_cp: begin
				cp_oe = 1;
				t1_we = 1;
			end
		
		`inc_cp + 'd1: begin
				t1_oe = 1;
				cp_we = 1;
				alu_oe = 1;
				alu_opcode = `INC;
			end
		
		`load_rg: begin
				regs_addr = rg;
				regs_oe = 1;
				if(d)
					t1_we = 1;
				else
					t2_we = 1;
			end
		
		`load_rm_reg: begin
				//TODO load operand from register
			end
		
		`load_rm_mem: begin
				//TODO load operand from computed address
			end
		
		`load_rm_mem + 'd1: begin
				//TODO load operand from computed address
			end
		
		`exec: begin
				t1_oe = 1;
				t2_oe = 1;
				alu_carry = ind[0];
				case(ri[6:4])
					7'b000: alu_opcode = `ADD;
					7'b100: alu_opcode = `ADC;
					7'b010: alu_opcode = `SUB;
					7'b110: alu_opcode = `SBB;
					7'b001: alu_opcode = `AND;
					7'b101: alu_opcode = `OR;
					7'b011: alu_opcode = `XOR;
				endcase
				alu_oe = 1;
				//TODO store operation result
				ind_sel = 1;
				ind_we = 1;
			end
		
		`exec + 'd1: begin
				//TODO store operation result
				end
		
		`addr_sum: begin
				//TODO compute and load address when RI[13] == 0
			end
		
		`addr_sum + 'd1: begin
				//TODO compute and load address when RI[13] == 0
			end
		
		`addr_sum + 'd2: begin
				//TODO compute and load address when RI[13] == 0
			end
		
		`addr_reg: begin
				//TODO compute and load address when RI[13] == 1
			end
		
		default: ;
	endcase
end

endmodule
